# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2019 Virtuozzo International GmbH
#

"""
God class fo working with VZ repository subscription service
"""
import sys
import os
import re
import datetime
import zipfile
import logging
import logging.config
import errno
import subprocess
import configparser
import traceback
import datetime
import requests
import platform
import json
import warnings

LOGFILE_PATH = u"/var/log/vzsubscribe/vzsubscribe.log"
EULA_ACCEPTED = u"/var/cache/vzsubscribe/eula_accepted"
CERT_PATH = u"/var/lib/vzsubscribe/certs"
EULA_FILE = "/usr/share/licenses/vzsubscribe/EULA"
VZS_CFG = u"/etc/vzsubscribe/vzsubscribe.ini"
VZS_STATE = u"/var/lib/vzsubscribe/state.ini"

DEFAULT_SERVER_URL = "https://rk.virtuozzo.com"
DEFAULT_SERVER_PORT = 4080
DEFAULT_REPO_NAME = "Virtuozzo ReadyKernel Patches"
DEFAULT_REPO_ID = "rk-patches"
#pylint: disable=W0107, C0103

class More(object):
    '''
    Class to print long text with "more"-like functionality
    '''
    def __init__(self, num_lines):
        self.num_lines = num_lines-1

    def __ror__(self, other):
        '''
        Pipe operator
        '''  
        s = str(other).split("\n")
        for i in xrange(0, len(s), self.num_lines):
            print("\n".join(s[i:i+self.num_lines]))
            raw_input("Press <Enter> for more")


ZERO = datetime.timedelta(0)

class UTC(datetime.tzinfo):
    '''
    Small ugly snippet to fix UTC-issue with python datetime conversion.
    '''
    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
       return ZERO

UTC_TZ = UTC()

class ZipAll(object):
    '''
    Lazy way to add files and folders to zip
    '''
    def __init__(self, zipname):
        '''
        Open Zipfilename
        '''
        self.zf = zipfile.ZipFile(zipname, 'w')
        pass

    def to_zip(self, filename2store):
        '''
        Add file or folder to zipfile
        '''
        if not os.path.exists(filename2store):
            return

        if os.path.isfile(filename2store):
            self.zf.write(filename2store)

        if os.path.isdir(filename2store):
            #pylint: disable=W0612
            for root, _dirnames, filenames in os.walk(filename2store):
                for filename in filenames:
                    filepath = os.path.realpath(os.path.join(root, filename))
                    self.zf.write(filepath)

    def close(self):
        '''
        Close ZIP file.
        '''
        self.zf.close()


def logmethod(method):
    """
    Decorator to log call end exists from methods
    """
    def _wrapper(self, *args, **kwds):
        "Use this one for instance or class methods"
        self.log.debug("BEGIN " + method.__name__)
        ret = method(self, *args, **kwds)
        self.log.debug("END " + method.__name__)
        return ret
    return _wrapper

def file2string(filepath):
    """
    Read file into string
    """
    lstr = None
    if os.path.exists(filepath):
        lfile = open(filepath, "r")
        lstr = lfile.read()
        lfile.close()
    return lstr


def str2file(str_, filename):
    '''
    Short writing given string to the file.
    '''
    lfile = open(filename, 'w')
    lfile.write(str_)
    lfile.close()

def mkdir_p(path):
    """
    Create path if not exists
    """
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
    pass

class VZSM(object):
    '''
    Business logic 
    '''
    def __init__(self):
        '''
        Getting base info about current SM subscription
        '''
        self.last_return_code = 0
        linux_distribution = platform.linux_distribution()
        # For cases when our binaries are located in custom path
        self.prefix = ""

        if os.path.isfile('/etc/debian_version'):
            self.repo_filepath = '/etc/apt/sources.list.d/subscription.list'
            self.is_deb = True
        else:
            self.repo_filepath = '/etc/yum.repos.d/rk-patch.repo'
            self.is_deb = False

        mkdir_p(os.path.dirname(LOGFILE_PATH))
        mkdir_p(os.path.dirname(EULA_ACCEPTED))
        mkdir_p(os.path.dirname(VZS_CFG))
        mkdir_p(os.path.dirname(VZS_STATE))
        mkdir_p(CERT_PATH)

        self.log = logging.getLogger("VZSM")
        self.log.setLevel(logging.DEBUG)
        handler = logging.handlers.RotatingFileHandler(LOGFILE_PATH, maxBytes=20*1024, backupCount=10)
        formatter = logging.Formatter('%(asctime)s:%(levelname)s:%(module)s> %(message)s')
        handler.setFormatter(formatter)
        self.log.addHandler(handler)

        if self.is_deb:
            yum_vars_dir = "/var/lib/sm/vars"
        else:
            yum_vars_dir = "/etc/yum/vars"
        mkdir_p(yum_vars_dir)

        self.vzs_cfg = configparser.ConfigParser()
        self.vzs_state = configparser.ConfigParser()

        try:
            self.vzs_cfg.read(VZS_CFG)
            self.vzs_state.read(VZS_STATE)
        except Exception:
            pass

        try:
            self.server = self.vzs_cfg['default']['server']
        except:
            self.server = DEFAULT_SERVER_URL

        try:
            self.server_port = self.vzs_cfg['default']['server_port']
        except:
            self.server_port = DEFAULT_SERVER_PORT

        try:
            self.repo_name = self.vzs_cfg['default']['repo_name']
        except:
            self.repo_name = DEFAULT_REPO_NAME

        try:
            self.repo_id = self.vzs_cfg['default']['repo_id']
            self.repo_filepath = '/etc/yum.repos.d/%s.repo' % (self.repo_id)
        except:
            self.repo_id = DEFAULT_REPO_ID

        try:
            self.register_attempt_date = self.vzs_state['default']['register_attempt_date']
        except:
            self.register_attempt_date = None

        try:
            self.registered_kakey = self.vzs_state['default']['registered_kakey']
        except:
            self.registered_kakey = None

        try:
            self.registered_hwid = self.vzs_state['default']['registered_hwid']
        except:
            self.registered_hwid = None

        osname_varfile = os.path.join(yum_vars_dir, "cpsm_osname")
        osversion_varfile = os.path.join(yum_vars_dir, "cpsm_osversion")

        #normalization
        dist_name = linux_distribution[0].lower()
        self.dist_name = '-'.join(dist_name.split()).replace("-linux", "")
        if self.dist_name == "virtuozzo" and os.path.exists("/etc/openvz-release"):
            self.dist_name = "openvz"

        if not os.path.exists(osname_varfile):
            with open(osname_varfile, "w") as lf_:
              lf_.write(self.dist_name)

        dist_major_version = linux_distribution[1].split('.')[0] if not self.is_deb else linux_distribution[1].replace('.', '-')
        self.dist_major_version = dist_major_version
        if not os.path.exists(osversion_varfile):
            with open(osversion_varfile, "w") as lf_:
              lf_.write(dist_major_version)

        pass


    @logmethod
    def get_output(self, scmd):
        '''
        Just get output of OS command.
        Log commands and results
        '''
        #pylint: disable=W0612
        self.log.debug(scmd)
        # progin, progout = os.popen4(scmd)

        proc = subprocess.Popen(scmd, shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                close_fds=True)
        out, _ = proc.communicate()
        proc.wait()
        self.last_return_code = proc.returncode
        self.log.debug(out)
        return out


    @logmethod
    def info(self, verbose=False):
        """
        Lot of info about state of subscription
        """
        def out_(msg):
            '''
            Print and log
            '''
            self.log.info(msg)
            print(msg)

        out_("=== Summary ===")
        self.repo_info()

        return True

    @logmethod
    def diagnostics(self):
        """
        Prepare diagnostics pack
        """
        #pylint: disable=E1101
        now = datetime.datetime.now(UTC_TZ).strftime('%Y-%m-%d-%H-%M-%S')
        diagnostics_filename = 'vzsubscribe-diagnostics-pack-' + now + '.zip'

        za = ZipAll(diagnostics_filename)

        for path in [LOGFILE_PATH,
                     '/var/log/yum.log']:
            try:
                za.to_zip(path)
            except:
                pass
        za.close()

        txtinfo = '''
    Prepared diagnostics pack
       %(diagnostics_filename)s.
    If you have any problems with subscriptions,
    please, send it to our support.
        ''' % vars()
        self.log.info(txtinfo)
        print(txtinfo)
        return True

    @logmethod
    def get_repo_urls(self):
        """
        Collect repositories from subscription repo.
        """
        repo_file_content = file2string(self.repo_filepath)
        self.log.debug(repo_file_content)
        repo_urls = set()
        if not repo_file_content:
           return repo_urls

        if self.is_deb:
            re_ = re.compile(r'(?m)deb \[arch=amd64\] (?P<url>\S+) .*$')
        else:
            re_ = re.compile(r'(?m)baseurl\s*=\s*(?P<url>.+)\s*$')

        for url in re_.findall(repo_file_content):
            repo_urls.add(url)
        return repo_urls

    @logmethod
    def repo_info(self):
        """
        Print info about subscriptions
        """
        repo_urls = self.get_repo_urls()
        if repo_urls:
            info = ''.join(['The system successfully registered, attached repositories:\n  *  ',
                            "\n  *  ".join(repo_urls)])
            self.log.info(info)
            print(info)
            return True
        return False

    @logmethod
    def get_key(self, uuid):
        proc = subprocess.Popen(['/usr/sbin/vz-get-rk', uuid, '1'],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    close_fds=True)
        out, _ = proc.communicate()
        proc.wait()
        if out:
            return out
        else:
            return uuid

    @logmethod
    def register(self, activation_key, eula_accepted=False):
        '''
        Register the system with given activation_key
        '''
        #pylint: disable=W0612, W0613

        if not eula_accepted and not os.path.exists(EULA_ACCEPTED) and os.path.exists(EULA_FILE):
            # eula_text  = ''

            from unidecode import unidecode
            import codecs
            eula_text = unidecode(
                codecs
                    .open(EULA_FILE, 'r', 'utf-8')
                    .read()
            ).encode('ascii')

            import shlex
            rows = 40
            try:
                rows_ = int(subprocess.check_output(shlex.split('tput lines')))
                if rows_:
                    rows  = rows_
            except:
                pass
            more = More(num_lines=rows)
            eula_text | more
            accept = ''
            while not accept:
              accept = raw_input("Accept EULA? [y/n]")
            if accept and accept[0] in ['y', 'Y']:
                eula_accepted = True

        if eula_accepted and not os.path.exists(EULA_ACCEPTED):
            with open(EULA_ACCEPTED, 'w') as lf_:
                lf_.write('EULA ACCEPTED')

        eula_accepted |=  os.path.exists(EULA_ACCEPTED)

        if not eula_accepted:
            print("EULA not accepted, can't continue")
            return False

        if 'els' in self.server and not os.path.exists('/proc/vz/hwid'):
            print("Can't find hwid")
            return False

        all_hwids = self.get_output("/usr/sbin/vz-get-hwid").split()
        if all_hwids and self.registered_hwid:
            hwid_index = all_hwids.index(self.registered_hwid)
            if hwid_index:
                all_hwids.insert(0, all_hwids.pop(hwid_index))

        today = datetime.datetime.now().strftime('%Y-%m-%d')
        kakey = None

        # Need to unify ELS and RK in future?
        if 'els' in self.server:
            for l in self.get_output("/usr/sbin/vzlicview").split():
                if "key_number" in l:
                    try:
                        kakey = l.split('"')[1]
                    except:
                        print("Failed to parse key")
                        kakey = None

            if not kakey:
                print("License key is not found")
                return False

            if self.register_attempt_date and self.register_attempt_date==today and self.registered_kakey==kakey:
                print("ELS is already registered today (remove '%s' file to re-register)" % VZS_STATE)
                return False

            reg_url = '%s:%s/register?kakey=%s&hwid=HWID_PLACEHOLDER' % \
                       (self.server, self.server_port, kakey)
            cert = '/usr/share/vzsubscribe/connect.pem'
        elif activation_key:
            reg_url = '%s:%s/register?os=%s%s&hwid=HWID_PLACEHOLDER&key=%s' % \
                       (self.server, self.server_port, self.dist_name, self.dist_major_version, activation_key)
            cert = None
        else:
            if not os.path.isfile("/usr/sbin/vzlicview"):
                print("No activation key specified and no vzlicview tool found, nothing to register")
                sys.exit(12)

            for l in self.get_output("/usr/sbin/vzlicview").split():
                if "key_number" in l:
                    try:
                        kakey = l.split('"')[1]
                    except:
                        print("Failed to parse VZ license key")
                        kakey = None
            if not kakey:
                print("Failed to get VZ license key to activate RK with it")
                sys.exit(13)
            reg_url = '%s:%s/register?os=virtuozzo7&kakey=%s&hwid=HWID_PLACEHOLDER' % \
                       (self.server, self.server_port, kakey)
            cert = '/usr/share/vzsubscribe/vz7rk.pem'

        register_success = False
        failure_msgs = []
        js_answer = {}
        hwid = None

        for hwid in all_hwids:
            reg_url_hwid = reg_url.replace("HWID_PLACEHOLDER", hwid)
            try:
                if cert:
                    r = requests.get(reg_url_hwid, verify=True, cert=cert)
                else:
                    r = requests.get(reg_url_hwid, verify=True)
            except Exception as e:
                try:
                    failure_msgs.append("  Failed to send registration request: " + str(e).replace(activation_key, "").replace(hwid, ""))
                except:
                    failure_msgs.append("  Failed to send registration request. Failed to reach %s:%s" % (self.server, self.server_port))
                    continue
                continue

            try:
                js_answer = json.loads(r.text)
                cert = js_answer['PrivateKey']
            except Exception as e:
                if r.status_code != 0 and r.status_code != 200:
                    if 'els' in self.server:
                        failure_msgs.append("  Wrong response to ELS repo registration request, return code: %s %s" % (r.status_code, r.text))
                    else:
                        failure_msgs.append("  Wrong response to repo registration request, return code: %s %s" % (r.status_code, r.text))

                    # ELS is not activated
                    if r.status_code == 500 and js_answer and 'Error' in js_answer:
                        if js_answer['Error'] == 'Please buy ELS support':
                            break
                else:
                    if 'els' in self.server:
                        failure_msgs.append("  Failed to parse ELS server response:\n" + str(e))
                    else:
                        failure_msgs.append("  Failed to parse server response:\n" + str(e))
                continue

            register_success = True
            break

        if not register_success:
            if failure_msgs:
                if 'els' in self.server:
                    print("Registration of ELS repo failed:")
                else:
                    print("Registration of RK repo failed:")
            else:
                if 'els' in self.server:
                    print("Registration of ELS repo failed for unknown reason")
                else:
                    print("Registration of RK repo failed for unknown reason")
            prev_msg = "";
            for msg in failure_msgs:
                if msg != prev_msg:
                    print(msg)
                    prev_msg = msg
            return False

        cert = js_answer['Cert']
        url_key = self.get_key(js_answer['PrivateKey'])
        with open(CERT_PATH + "/vz-rk.pem", 'w') as cf_:
            cf_.write(cert)

        with open(self.repo_filepath, 'w') as rf_:
            if self.is_deb:
                print("Not supported yet")
            else:
                rf_.write('''[%s]
name=%s
baseurl=%s/%s/
enabled=1
gpgcheck=0
sslverify=1
priority=50
rk_repo=1
skip_if_unavailable=1
sslclientcert=%s/vz-rk.pem
''' % (self.repo_id, self.repo_name, self.server, url_key, CERT_PATH))

        is_ok = self.repo_info()
        if not is_ok:
            info = '\n'.join([
                'Troubles with registration, no repository attached.',
                'If problem persists, call me with "diagnostics" options'
                ])
            self.log.error(info)
            self.log.error("HWID=" + hwid)
            self.log.error("activation_key=" + activation_key)
            print(info)
            return False

        # Save successful registration info
        if 'els' in self.server:
            self.register_attempt_date = today
            self.registered_kakey = kakey
            self.registered_hwid = hwid

            with warnings.catch_warnings(record=True) as w:
                # Hide configparser internal deprecation warnings
                warnings.simplefilter("ignore")

                if 'default' not in self.vzs_state.sections():
                    self.vzs_state.add_section('default')

                self.vzs_state.set('default', 'register_attempt_date', today)
                self.vzs_state.set('default', 'registered_kakey', kakey)
                self.vzs_state.set('default', 'registered_hwid', hwid)

            try:
                with open(VZS_STATE, 'w') as configfile:
                    self.vzs_state.write(configfile)
            except:
                pass

        return True

    @logmethod
    def unregister(self, silent=False):
        '''
        Unregistering current HWID
        '''
        hwid = self.get_output("/usr/sbin/vz-get-hwid").split()[0]
        res = ""
        info = ""

        # "Unregister" for ELS means clean up of a local cache
        # and removal of els.repo file.
        # vzsubscribe can't be used to drop ELS flag from the key in KA
        if 'els' in self.server:
            if os.path.isfile(VZS_STATE):
                try:
                    os.remove(VZS_STATE)
                except Exception as e:
                    print("Failed to remove state file %s : %s" % (VZS_STATE, str(e)))
            info = "Information about ELS repository has been removed from the node. Note that ELS feature is still assigned to your Vz key."
            print(info)
        else:
            reg_url = '%s:4080/unregister?hwid=%s' % (self.server, hwid)
            try:
                r = requests.get(reg_url, verify=True)
            except Exception as e:
                print("Failed to send unregister request")
                print(str(e).replace(hwid, ""))
                return False

            if r.status_code != 0:
                print("Wrong response to unregister request, return code: %s" % (r.status_code))
                print(r.text)
                return False
            res = r.text

            if 'has been unregistered' in res:
                info = 'Successfully unregistered'
            elif 'is currently not registered' in res:
                info = 'The system already unregistered'
            else:
                info = "Troubles with deregistration. "

        self.log.info(info)

        for file_ in os.listdir(CERT_PATH):
            filename = os.path.join(CERT_PATH, file_)
            try:
                os.remove(filename)
            except Exception as e:
                print("Failed to remove certificate file %s : %s" % (filename, str(e)))

        if os.path.exists(self.repo_filepath):
            try:
                os.remove(self.repo_filepath)
            except Exception as e:
                print("Failed to remove repository file %s : %s" % (self.repo_filepath, str(e)))

            if not (silent and 'already' in info):
                print(info)
        return True


if __name__ == '__main__':
    vzsm = VZSM()
    vzsm.diagnostics()
    pass
