# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2019 Virtuozzo International GmbH
#
"""
Provides entry point main().
"""
__version__ = "2.0.0"

import sys
import os
import argparse
import traceback
import VZSM

def patched_error(self, message):
    """error(message: string)

    Prints a usage message incorporating the message to stderr and
    exits.
    """
    self.print_usage(sys.stderr)
    msg = '%s: error: %s\n' % (self.prog, message)
    try:
        print(msg.decode('string_escape'))
    except:
        print(msg)
    self.exit(os.EX_USAGE)

argparse.ArgumentParser.error = patched_error

class AliasedSubParsersAction(argparse._SubParsersAction):
    """
    Hack to support aliases in arparse before python 3.0.
    """
    class _AliasedPseudoAction(argparse.Action):
        def __init__(self, name, aliases, help):
            dest = name
            if aliases:
                dest += ' (%s)' % ','.join(aliases)
            sup = super(AliasedSubParsersAction._AliasedPseudoAction, self)
            sup.__init__(option_strings=[], dest=dest, help=help)

    def add_parser(self, name, **kwargs):
        if 'aliases' in kwargs:
            aliases = kwargs['aliases']
            del kwargs['aliases']
        else:
            aliases = []

        parser = super(AliasedSubParsersAction, self).add_parser(name, **kwargs)

        # Make the aliases work.
        for alias in aliases:
            self._name_parser_map[alias] = parser
        # Make the help text reflect them, first removing old help entry.
        if 'help' in kwargs:
            help = kwargs.pop('help')
            self._choices_actions.pop()
            pseudo_action = self._AliasedPseudoAction(name, aliases, help)
            self._choices_actions.append(pseudo_action)

        return parser


class VZlicenser(object):
    """
    Main class for all commands
    """
    def __init__(self):
        """Set up and parse command line options"""
        parser = argparse.ArgumentParser(prog='vzsubscribe')
        parser.register('action', 'parsers', AliasedSubParsersAction)

        parser.add_argument('--verbose', action='store_true', dest='verbose', default=False,
                            help='Log a lot of debug info')


        subparsers = parser.add_subparsers(dest='command', help='sub-command help')
        parser_register = subparsers.add_parser('register', help='init help', aliases=['r'])
        parser_register.add_argument('--accept-eula', action='store_true', dest='accept_eula', default=False,
                            help='Automatically accept EULA')

        parser_register.add_argument('org', nargs='?', help='Your Organization')
        parser_register.add_argument('key', nargs='?', help='Activation Key')

        parser_info = subparsers.add_parser('info',
                                                help='Information about subscriptions',
                                                aliases=['i'])

        parser_diagnostics = subparsers.add_parser('diagnostics',
                                                help='Make diagnostics pack for troubleshooting',
                                                aliases=['d'])

        parser_unregister = subparsers.add_parser('unregister',
                                                help='Unregister system',
                                                aliases=['u'])

        parser_heal = subparsers.add_parser(
            'heal',
            help='Dummy option, provided for backward compatibility',
            aliases=['h'])

        parser_heal.add_argument('--force', action='store_true', dest='force', default=False,
                            help='Force healing now')

        try:
            self.args = parser.parse_args()
        except Exception as ex:
            sys.exit(os.EX_USAGE)

        self.vzsm = VZSM.VZSM()


    def process(self):
        """
        Main entry point.
        Process command line options, call appropriate logic.
        """
        if self.args.command in ['register', 'r']:
            eula_accepted = False
            eula_accepted |= self.args.accept_eula
            eula_accepted |= os.path.exists('/etc/vzlinux-release')

            # org is left for backward compatibility with readykernel tool
            org = self.args.org
            key = self.args.key
            if not org and key:
                #One parameter, getting org from key
                terms = []
                for split_key in [':', '$', '@']:
                    terms = self.args.key.split(split_key)
                    if len(terms) == 2:
                        org, key = terms
                        break
            if not key and org:
                key = org
            res = self.vzsm.register(key, eula_accepted)
            return res 

        if self.args.command in ['heal', 'h']:
            # For backward compatibility
            return True

        if self.args.command in ['info', 'i']: 
            return self.vzsm.info(verbose=self.args.verbose)

        if self.args.command in ['diagnostics', 'd']:
            return self.vzsm.diagnostics()

        if self.args.command in ['unregister', 'u']:
            return self.vzsm.unregister()

        pass


def main():
    """
     Start procedure
    """
    if os.geteuid() != 0: 
        print("You have to run it as root!")
        sys.exit(os.EX_USAGE)

    res = False

    try:
        vl = VZlicenser()
        res = vl.process()
    except KeyboardInterrupt:
        print('Interrupted by user keyboard!')
        sys.exit(os.EX_USAGE)

    if not res:
        sys.exit(os.EX_UNAVAILABLE)
    sys.exit(os.EX_OK)


if __name__ == '__main__':
    res = main()
    sys.exit(os.EX_OK)

